@@ -3,36 +3,55 @@
 '''
 A library to access Tryton's models like a client.
 '''
-__version__ = "2.2.1"
-__all__ = ['Model', 'Wizard']
-from types import NoneType
+__version__ = "3.4.0"
+__all__ = ['Model', 'Wizard', 'Report']
+import sys
+try:
+    import cdecimal
+    # Use cdecimal globally
+    if 'decimal' not in sys.modules:
+        sys.modules['decimal'] = cdecimal
+except ImportError:
+    import decimal
+    sys.modules['cdecimal'] = decimal
 import threading
 import datetime
 from decimal import Decimal
+from types import NoneType
+
 import proteus.config
-from proteus.pyson import PYSONDecoder
 
 _MODELS = threading.local()
 
+
 class _EvalEnvironment(dict):
     'Dictionary for evaluation'
-    def __init__(self, parent):
+    def __init__(self, parent, eval_type='eval'):
         super(_EvalEnvironment, self).__init__()
         self.parent = parent
+        assert eval_type in ('eval', 'on_change')
+        self.eval_type = eval_type
 
     def __getitem__(self, item):
-        if item == '_parent_' + self.parent._parent_field_name \
-                and self.parent.parent:
-            return _EvalEnvironment(self.parent.parent)
-        return self.parent._get_eval()[item]
+        if item == '_parent_' + self.parent._parent_name \
+                and self.parent._parent:
+            return _EvalEnvironment(self.parent._parent,
+                eval_type=self.eval_type)
+        if self.eval_type == 'eval':
+            return self.parent._get_eval()[item]
+        else:
+            return self.parent._get_on_change_values(fields=[item])[item]
 
     def __getattr__(self, item):
-        return self.__getitem__(item)
+        try:
+            return self.__getitem__(item)
+        except KeyError:
+            raise AttributeError(item)
 
     def get(self, item, default=None):
         try:
-            return self.__getattr__(item)
-        except:
+            return self.__getitem__(item)
+        except KeyError:
             pass
         return super(_EvalEnvironment, self).get(item, default)
 
@@ -45,7 +64,13 @@ class _EvalEnvironment(dict):
     __repr__ = __str__
 
     def __contains__(self, item):
-        return item in self.parent._fields
+        if item == '_parent_' + self.parent._parent_name \
+                and self.parent._parent:
+            return True
+        if self.eval_type == 'eval':
+            return item in self.parent._get_eval()
+        else:
+            return item in self.parent._fields
 
 
 class FieldDescriptor(object):
@@ -58,12 +83,12 @@ class FieldDescriptor(object):
         self.__doc__ = definition['string']
 
     def __get__(self, instance, owner):
-        if instance.id > 0 and self.name not in instance._values:
+        if instance.id > 0:
             instance._read(self.name)
         return instance._values.get(self.name, self.default)
 
     def __set__(self, instance, value):
-        if instance.id > 0 and self.name not in instance._values:
+        if instance.id > 0:
             instance._read(self.name)
         previous = getattr(instance, self.name)
         instance._values[self.name] = value
@@ -87,7 +112,7 @@ class CharDescriptor(FieldDescriptor):
     default = None
 
     def __set__(self, instance, value):
-        assert isinstance(value, basestring) or value in (None, False)
+        assert isinstance(value, basestring) or value is None
         super(CharDescriptor, self).__set__(instance, value or '')
 
 
@@ -95,30 +120,30 @@ class BinaryDescriptor(FieldDescriptor):
     default = None
 
     def __set__(self, instance, value):
-        assert isinstance(value, (basestring, buffer)) or value in (None, False)
-        super(BinaryDescriptor, self).__set__(instance, value or '')
+        assert isinstance(value, (basestring, buffer)) or value is None
+        super(BinaryDescriptor, self).__set__(instance, value)
 
 
 class IntegerDescriptor(FieldDescriptor):
-    default = 0
 
     def __set__(self, instance, value):
-        assert isinstance(value, (int, long))
+        assert isinstance(value, (int, long, NoneType))
         super(IntegerDescriptor, self).__set__(instance, value)
 
 
 class FloatDescriptor(FieldDescriptor):
-    default = 0.0
 
     def __set__(self, instance, value):
-        super(FloatDescriptor, self).__set__(instance, float(value))
+        assert isinstance(value, (int, long, float, Decimal, NoneType))
+        if value is not None:
+            value = float(value)
+        super(FloatDescriptor, self).__set__(instance, value)
 
 
 class NumericDescriptor(FieldDescriptor):
-    default = Decimal('0.0')
 
     def __set__(self, instance, value):
-        assert isinstance(value, Decimal)
+        assert isinstance(value, (NoneType, Decimal))
         # TODO add digits validation
         super(NumericDescriptor, self).__set__(instance, value)
 
@@ -139,7 +164,6 @@ class ReferenceDescriptor(FieldDescriptor):
         if isinstance(value, basestring):
             assert value.startswith(',')
         elif isinstance(value, Model):
-            assert value.id > 0 and not value._changed
             assert value._config == instance._config
         super(ReferenceDescriptor, self).__set__(instance, value)
 
@@ -153,31 +177,41 @@ class DateDescriptor(FieldDescriptor):
         return value
 
     def __set__(self, instance, value):
-        assert isinstance(value, datetime.date) or value in (None, False)
+        assert isinstance(value, datetime.date) or value is None
         super(DateDescriptor, self).__set__(instance, value)
 
 
 class DateTimeDescriptor(FieldDescriptor):
     def __set__(self, instance, value):
-        assert isinstance(value, datetime.datetime) or value in (None, False)
+        assert isinstance(value, datetime.datetime) or value is None
         super(DateTimeDescriptor, self).__set__(instance, value)
 
 
+class TimeDescriptor(FieldDescriptor):
+    def __set__(self, instance, value):
+        assert isinstance(value, datetime.time) or value is None
+        super(TimeDescriptor, self).__set__(instance, value)
+
+
+class DictDescriptor(FieldDescriptor):
+    def __set__(self, instance, value):
+        assert isinstance(value, dict) or value is None
+        super(DictDescriptor, self).__set__(instance, value)
+
+
 class Many2OneDescriptor(FieldDescriptor):
     def __get__(self, instance, owner):
         relation = Model.get(self.definition['relation'], instance._config)
         value = super(Many2OneDescriptor, self).__get__(instance, owner)
-        if isinstance(value, (int, long)) and value is not False:
+        if isinstance(value, (int, long)):
             value = relation(value)
-        elif not value:
-            value = None
-        instance._values[self.name] = value
+        if self.name in instance._values:
+            instance._values[self.name] = value
         return value
 
     def __set__(self, instance, value):
         assert isinstance(value, (Model, NoneType))
         if value:
-            assert value.id > 0 and not value._changed
             assert value._config == instance._config
         super(Many2OneDescriptor, self).__set__(instance, value)
 
@@ -190,12 +224,17 @@ class One2ManyDescriptor(FieldDescriptor):
     default = []
 
     def __get__(self, instance, owner):
+        from .pyson import PYSONDecoder
         relation = Model.get(self.definition['relation'], instance._config)
         value = super(One2ManyDescriptor, self).__get__(instance, owner)
         if not isinstance(value, ModelList):
-            value = ModelList((relation(id)
-                for id in value or []), instance, self.name,
-                self.definition.get('relation_field'))
+            ctx = instance._context.copy() if instance._context else {}
+            if self.definition.get('context'):
+                decoder = PYSONDecoder(_EvalEnvironment(instance))
+                ctx.update(decoder.decode(self.definition.get('context')))
+            with instance._config.set_context(ctx):
+                value = ModelList(self.definition, (relation(id)
+                        for id in value or []), instance, self.name)
             instance._values[self.name] = value
         return value
 
@@ -222,13 +261,13 @@ class ReferenceValueDescriptor(ValueDescriptor):
         value = super(ReferenceValueDescriptor, self).__get__(instance, owner)
         if isinstance(value, Model):
             value = '%s,%s' % (value.__class__.__name__, value.id)
-        return value or False
+        return value or None
 
 
 class Many2OneValueDescriptor(ValueDescriptor):
     def __get__(self, instance, owner):
         value = super(Many2OneValueDescriptor, self).__get__(instance, owner)
-        return value and value.id or False
+        return value and value.id or None
 
 
 class One2OneValueDescriptor(Many2OneValueDescriptor):
@@ -237,18 +276,31 @@ class One2OneValueDescriptor(Many2OneValueDescriptor):
 
 class One2ManyValueDescriptor(ValueDescriptor):
     def __get__(self, instance, owner):
-        value = [('add', [])]
+        value = []
         value_list = getattr(instance, self.name)
+        parent_name = self.definition.get('relation_field', '')
+        to_add = []
+        to_create = []
+        to_write = []
         for record in value_list:
-            if record.id > 0:
-                if record._changed:
-                    value.append(('write', record.id, record._get_values(
-                        fields=record._changed)))
-                value[0][1].append(record.id)
+            if record.id >= 0:
+                values = record._get_values(fields=record._changed)
+                values.pop(parent_name, None)
+                if record._changed and values:
+                    to_write.extend(([record.id], values))
+                to_add.append(record.id)
             else:
-                value.append(('create', record._get_values()))
+                values = record._get_values()
+                values.pop(parent_name, None)
+                to_create.append(values)
+        if to_add:
+            value.append(('add', to_add))
+        if to_create:
+            value.append(('create', to_create))
+        if to_write:
+            value.append(('write',) + tuple(to_write))
         if value_list.record_removed:
-            value.append(('unlink', [x.id for x in value_list.record_removed]))
+            value.append(('remove', [x.id for x in value_list.record_removed]))
         if value_list.record_deleted:
             value.append(('delete', [x.id for x in value_list.record_deleted]))
         return value
@@ -273,7 +325,7 @@ class ReferenceEvalDescriptor(EvalDescriptor):
         value = super(ReferenceEvalDescriptor, self).__get__(instance, owner)
         if isinstance(value, Model):
             value = '%s,%s' % (value.__class__.__name__, value.id)
-        return value or False
+        return value or None
 
 
 class Many2OneEvalDescriptor(EvalDescriptor):
@@ -281,7 +333,7 @@ class Many2OneEvalDescriptor(EvalDescriptor):
         value = super(Many2OneEvalDescriptor, self).__get__(instance, owner)
         if value:
             return value.id
-        return False
+        return None
 
 
 class One2OneEvalDescriptor(Many2OneEvalDescriptor):
@@ -290,7 +342,14 @@ class One2OneEvalDescriptor(Many2OneEvalDescriptor):
 
 class One2ManyEvalDescriptor(EvalDescriptor):
     def __get__(self, instance, owner):
-        return [x.id for x in getattr(instance, self.name)]
+        # Directly use _values to prevent infinite recursion with
+        # One2ManyDescriptor which could evaluate this field to decode the
+        # context
+        value = instance._values.get(self.name, [])
+        if isinstance(value, ModelList):
+            return [x.id for x in value]
+        else:
+            return value
 
 
 class Many2ManyEvalDescriptor(One2ManyEvalDescriptor):
@@ -302,9 +361,8 @@ class MetaModelFactory(object):
         'boolean': BooleanDescriptor,
         'char': CharDescriptor,
         'text': CharDescriptor,
-        'sha': CharDescriptor,
         'binary': BinaryDescriptor,
-        'selection': CharDescriptor, # TODO implement its own descriptor
+        'selection': CharDescriptor,  # TODO implement its own descriptor
         'integer': IntegerDescriptor,
         'biginteger': IntegerDescriptor,
         'float': FloatDescriptor,
@@ -313,6 +371,9 @@ class MetaModelFactory(object):
         'reference': ReferenceDescriptor,
         'date': DateDescriptor,
         'datetime': DateTimeDescriptor,
+        'timestamp': DateTimeDescriptor,
+        'time': TimeDescriptor,
+        'dict': DictDescriptor,
         'many2one': Many2OneDescriptor,
         'one2many': One2ManyDescriptor,
         'many2many': Many2ManyDescriptor,
@@ -339,9 +400,10 @@ class MetaModelFactory(object):
         self.config = config or proteus.config.get_config()
 
     def __call__(self):
-        models_key = 'c%s' % id(self.config)
+        models_key = 'c%su%s' % (id(self.config), self.config.user)
         if not hasattr(_MODELS, models_key):
             setattr(_MODELS, models_key, {})
+
         class MetaModel(type):
             'Meta class for Model'
             def __new__(mcs, name, bases, dict):
@@ -356,7 +418,8 @@ class MetaModelFactory(object):
                 for field_name, definition in dict['_fields'].iteritems():
                     if field_name == 'id':
                         continue
-                    Descriptor = self.descriptors[definition['type']]
+                    Descriptor = self.descriptors.get(definition['type'],
+                        FieldDescriptor)
                     dict[field_name] = Descriptor(field_name, definition)
                     VDescriptor = self.value_descriptors.get(
                             definition['type'], ValueDescriptor)
@@ -378,22 +441,26 @@ class MetaModelFactory(object):
 class ModelList(list):
     'List for Model'
 
-    def __init__(self, sequence=None, parent=None, parent_field_name='',
-            parent_name=''):
+    def __init__(self, definition, sequence=None, parent=None,
+            parent_field_name=''):
+        self.model_name = definition['relation']
         if sequence is None:
             sequence = []
         self.parent = parent
         if parent:
             assert parent_field_name
         self.parent_field_name = parent_field_name
-        self.parent_name = parent_name
+        self.parent_name = definition.get('relation_field', '')
+        self.domain = definition.get('domain', [])
+        self.context = definition.get('context')
+        self.add_remove = definition.get('add_remove')
         self.record_removed = set()
         self.record_deleted = set()
         result = super(ModelList, self).__init__(sequence)
         for record in self:
             record._parent = parent
             record._parent_field_name = parent_field_name
-            record._parent_name = parent_name
+            record._parent_name = self.parent_name
         return result
     __init__.__doc__ = list.__init__.__doc__
 
@@ -403,27 +470,16 @@ class ModelList(list):
             self.parent._changed.add(self.parent_field_name)
             self.parent._on_change(self.parent_field_name)
 
-    def append(self, record):
-        assert isinstance(record, Model)
-        if self.parent:
-            assert record._config == self.parent._config
-        elif self:
-            assert record._config == self[0]._config
-        assert record._parent is None
-        assert not record._parent_field_name
-        assert not record._parent_name
-        record._parent = self.parent
-        record._parent_field_name = self.parent_field_name
-        record._parent_name = self.parent_name
-        res = super(ModelList, self).append(record)
-        self._changed()
-        return res
-    append.__doc__ = list.append.__doc__
+    def _get_context(self):
+        from .pyson import PYSONDecoder
+        decoder = PYSONDecoder(_EvalEnvironment(self.parent))
+        ctx = self.parent._context.copy() if self.parent._context else {}
+        ctx.update(decoder.decode(self.context) if self.context else {})
+        return ctx
 
-    def extend(self, iterable):
-        iterable = list(iterable)
+    def __check(self, records):
         config = None
-        for record in iterable:
+        for record in records:
             assert isinstance(record, Model)
             if self.parent:
                 assert record._config == self.parent._config
@@ -433,13 +489,30 @@ class ModelList(list):
                 assert record._config == config
             else:
                 config = record._config
-        for record in iterable:
+        for record in records:
             assert record._parent is None
             assert not record._parent_field_name
             assert not record._parent_name
             record._parent = self.parent
             record._parent_field_name = self.parent_field_name
             record._parent_name = self.parent_name
+
+            # Set parent field to trigger on_change
+            if self.parent and self.parent_name in record._fields:
+                definition = record._fields[self.parent_name]
+                if definition['type'] in ('many2one', 'reference'):
+                    setattr(record, self.parent_name, self.parent)
+
+    def append(self, record):
+        self.__check([record])
+        res = super(ModelList, self).append(record)
+        self._changed()
+        return res
+    append.__doc__ = list.append.__doc__
+
+    def extend(self, iterable):
+        iterable = list(iterable)
+        self.__check(iterable)
         res = super(ModelList, self).extend(iterable)
         self._changed()
         return res
@@ -452,8 +525,8 @@ class ModelList(list):
     def pop(self, index=-1):
         self.record_removed.add(self[index])
         self[index]._parent = None
-        self[index]._parent_field_name = None
-        self[index]._parent_name = None
+        self[index]._parent_field_name = ''
+        self[index]._parent_name = ''
         res = super(ModelList, self).pop(index)
         self._changed()
         return res
@@ -462,8 +535,8 @@ class ModelList(list):
     def remove(self, record):
         self.record_deleted.add(record)
         record._parent = None
-        record._parent_field_name = None
-        record._parent_name = None
+        record._parent_field_name = ''
+        record._parent_name = ''
         res = super(ModelList, self).remove(record)
         self._changed()
         return res
@@ -477,6 +550,28 @@ class ModelList(list):
         raise NotImplementedError
     sort.__doc__ = list.sort.__doc__
 
+    def new(self, **kwargs):
+        'Adds a new record to the ModelList and returns it'
+        Relation = Model.get(self.model_name, self.parent._config)
+        with Relation._config.set_context(self._get_context()):
+            new_record = Relation(**kwargs)
+        self.append(new_record)
+        return new_record
+
+    def find(self, condition=None, offset=0, limit=None, order=None):
+        'Returns records matching condition taking into account list domain'
+        from .pyson import PYSONDecoder
+        decoder = PYSONDecoder(_EvalEnvironment(self.parent))
+        Relation = Model.get(self.model_name, self.parent._config)
+        if condition is None:
+            condition = []
+        field_domain = decoder.decode(self.domain)
+        add_remove_domain = (decoder.decode(self.add_remove)
+            if self.add_remove else [])
+        new_domain = [field_domain, add_remove_domain, condition]
+        with Relation._config.set_context(self._get_context()):
+            return Relation.find(new_domain, offset, limit, order)
+
 
 class Model(object):
     'Model class for Tryton records'
@@ -486,19 +581,20 @@ class Model(object):
     _config = None
     _fields = None
 
-    def __init__(self, id=None, **kwargs):
+    def __init__(self, id=None, _default=True, **kwargs):
         super(Model, self).__init__()
         if id:
             assert not kwargs
         self.__id = id or Model.__counter
         if self.__id < 0:
             Model.__counter -= 1
-        self._values = {} # store the values of fields
-        self._changed = set() # store the changed fields
-        self._parent = None # store the parent record
-        self._parent_field_name = '' # store the field name in parent record
-        self._parent_name = '' # store the field name to parent record
-        if self.id < 0:
+        self._values = {}  # store the values of fields
+        self._changed = set()  # store the changed fields
+        self._parent = None  # store the parent record
+        self._parent_field_name = ''  # store the field name in parent record
+        self._parent_name = ''  # store the field name to parent record
+        self._context = self._config.context  # store the context
+        if self.id < 0 and _default:
             self._default_get()
 
         for field_name, value in kwargs.iteritems():
@@ -510,17 +606,18 @@ class Model(object):
                 getattr(self, field_name).extend(value)
             else:
                 if definition['type'] == 'many2one':
-                    if (isinstance(value, (int, long)) and value is not False):
+                    if isinstance(value, (int, long)):
                         relation = Model.get(definition['relation'])
                         value = relation(value)
-                    elif value is False:
-                        value = None
                 setattr(self, field_name, value)
     __init__.__doc__ = object.__init__.__doc__
 
     @classmethod
     def get(cls, name, config=None):
         'Get a class for the named Model'
+        if isinstance(name, unicode):
+            name = name.encode('utf-8')
+
         class Spam(Model):
             __metaclass__ = MetaModelFactory(name, config=config)()
         return Spam
@@ -529,7 +626,7 @@ class Model(object):
     def reset(cls, config=None, *names):
         'Reset class definition for Models named'
         config = config or proteus.config.get_config()
-        models_key = 'c%s' % id(config)
+        models_key = 'c%su%s' % (id(config), config.user)
         if not names:
             setattr(_MODELS, models_key, {})
         else:
@@ -552,7 +649,8 @@ class Model(object):
     def __cmp__(self, other):
         'Compare with other'
         if isinstance(other, Model):
-            return cmp(self.id, other.id)
+            return cmp((self.__class__.__name__, self.id),
+                (other.__class__.__name__, other.id))
         if isinstance(other, (bool, NoneType)):
             return 1
         raise NotImplementedError
@@ -580,14 +678,14 @@ class Model(object):
         'Save the record'
         context = self._config.context
         if self.id < 0:
-            values = self._get_values(fields=self._changed)
-            self.__id = self._proxy.create(values, context)
+            values = self._get_values()
+            self.__id, = self._proxy.create([values], context)
         else:
             if not self._changed:
                 return
             values = self._get_values(fields=self._changed)
             context['_timestamp'] = self._get_timestamp()
-            self._proxy.write(self.id, values, context)
+            self._proxy.write([self.id], values, context)
         self.reload()
 
     def delete(self):
@@ -595,10 +693,23 @@ class Model(object):
         if self.id > 0:
             context = self._config.context
             context['_timestamp'] = self._get_timestamp()
-            return self._proxy.delete(self.id, context)
+            return self._proxy.delete([self.id], context)
         self.reload()
         return True
 
+    @classmethod
+    def duplicate(cls, records, default=None):
+        'Duplicate the record'
+        ids = cls._proxy.copy([r.id for r in records], default,
+            cls._config.context)
+        return [cls(id) for id in ids]
+
+    def click(self, button):
+        'Click on button'
+        self.save()
+        self.reload()  # Force reload because save doesn't always
+        return getattr(self._proxy, button)([self.id], self._config.context)
+
     def _get_values(self, fields=None):
         'Return dictionary values'
         if fields is None:
@@ -625,14 +736,17 @@ class Model(object):
 
     def _read(self, name):
         'Read field'
-        fields = []
-        if not self._values:
+        fields = [name]
+        if name in self._values:
+            return
+        loading = self._fields[name]['loading']
+        if loading == 'eager':
             fields = [x for x, y in self._fields.iteritems()
-                    if y['type'] not in ('one2many', 'many2many', 'binary')]
+                    if y['loading'] == 'eager']
+        if not self._fields:
             fields.append('_timestamp')
-        fields.append(name)
-        self._values.update(self._proxy.read(self.id, fields,
-            self._config.context))
+        self._values.update(self._proxy.read([self.id], fields,
+            self._config.context)[0])
         for field in fields:
             if (field in self._fields
                     and self._fields[field]['type'] == 'float'
@@ -647,7 +761,10 @@ class Model(object):
             self._config.context))
 
     def _default_set(self, values):
+        fieldnames = []
         for field, value in values.iteritems():
+            if '.' in field:
+                continue
             definition = self._fields[field]
             if definition['type'] in ('one2many', 'many2many'):
                 if value and len(value) and isinstance(value[0], (int, long)):
@@ -659,10 +776,13 @@ class Model(object):
                     record = relation()
                     record._default_set(vals)
                     records.append(record)
-                self._values[field] = ModelList(records, self, field,
-                        definition.get('relation_field', ''))
+                self._values[field] = ModelList(definition, records, self,
+                    field)
             else:
                 self._values[field] = value
+            fieldnames.append(field)
+        for field in sorted(fieldnames):
+            self._on_change(field)
 
     def _get_eval(self):
         values = dict((x, getattr(self, '__%s_eval' % x))
@@ -670,32 +790,53 @@ class Model(object):
         values['id'] = self.id
         return values
 
+    def _get_on_change_values(self, skip=None, fields=None):
+        values = {'id': self.id}
+        if fields:
+            definitions = ((f, self._fields[f]) for f in fields)
+        else:
+            definitions = self._fields.iteritems()
+        for field, definition in definitions:
+            if not fields:
+                if field == 'id' or (skip and field in skip):
+                    continue
+                if (self.id >= 0
+                        and (field not in self._values
+                            or field not in self._changed)):
+                    continue
+            if definition['type'] == 'one2many':
+                values[field] = [x._get_on_change_values(
+                        skip={definition.get('relation_field', '')})
+                    for x in getattr(self, field)]
+            elif (definition['type'] in ('many2one', 'reference')
+                    and self._parent_name == definition['name']
+                    and self._parent):
+                values[field] = self._parent._get_on_change_values(
+                    skip={self._parent_field_name})
+                if definition['type'] == 'reference':
+                    values[field] = (
+                        self._parent.__class__.__name__, values[field])
+            else:
+                values[field] = getattr(self, '__%s_eval' % field)
+        return values
+
     def _on_change_args(self, args):
         res = {}
-        values = self._get_eval()
-        del values['id']
-        for field, definition in self._fields.iteritems():
-            if definition['type'] in ('one2many', 'many2many'):
-                values[field] = [x._get_eval() for x in getattr(self, field)]
-        if self._parent:
-            values['_parent_%s' % self._parent_name] = \
-                    _EvalEnvironment(self._parent)
+        values = _EvalEnvironment(self, 'on_change')
         for arg in args:
             scope = values
             for i in arg.split('.'):
                 if i not in scope:
-                    scope = False
                     break
                 scope = scope[i]
-            res[arg] = scope
+            else:
+                res[arg] = scope
+        res['id'] = self.id
         return res
 
     def _on_change_set(self, field, value):
-        if self._fields[field]['type'] in ('one2many', 'many2many'):
-            if isinstance(value, (list, tuple)):
-                self._values[field] = value
-                self._changed.add(field)
-                return
+        if (self._fields[field]['type'] in ('one2many', 'many2many')
+                and not isinstance(value, (list, tuple))):
             to_remove = []
             if value and value.get('remove'):
                 for record_id in value['remove']:
@@ -706,29 +847,38 @@ class Model(object):
                 # remove without signal
                 list.remove(getattr(self, field), record)
             if value and value.get('add') or value.get('update'):
-                for vals in value.get('add', []):
+                for index, vals in value.get('add', []):
                     relation = Model.get(self._fields[field]['relation'],
                             self._config)
-                    record = relation()
+                    record = relation(_default=False)
                     for i, j in vals.iteritems():
+                        if '.' in i:
+                            continue
                         record._values[i] = j
                         record._changed.add(i)
                     # append without signal
-                    list.append(getattr(self, field), record)
+                    if index == -1:
+                        list.append(getattr(self, field), record)
+                    else:
+                        list.insert(getattr(self, field), index, record)
                 for vals in value.get('update', []):
                     if 'id' not in vals:
                         continue
                     for record in getattr(self, field):
                         if record.id == vals['id']:
                             for i, j in vals.iteritems():
+                                if '.' in i:
+                                    continue
                                 record._values[i] = j
                                 record._changed.add(i)
         else:
             self._values[field] = value
-            self._changed.add(field)
+        self._changed.add(field)
 
     def _on_change(self, name):
         'Call on_change for field'
+        # Import locally to not break installation
+        from proteus.pyson import PYSONDecoder
         definition = self._fields[name]
         if definition.get('on_change'):
             if isinstance(definition['on_change'], basestring):
@@ -747,6 +897,9 @@ class Model(object):
                 self._on_change_set(field, value)
             for field, value in later.iteritems():
                 self._on_change_set(field, value)
+        values = {}
+        to_change = set()
+        later = set()
         for field, definition in self._fields.iteritems():
             if not definition.get('on_change_with'):
                 continue
@@ -754,11 +907,23 @@ class Model(object):
                 continue
             if field == name:
                 continue
-            args = self._on_change_args(definition['on_change_with'])
+            if to_change & set(definition['on_change_with']):
+                later.add(field)
+            to_change.add(field)
+            values.update(self._on_change_args(definition['on_change_with']))
+        if to_change:
             context = self._config.context
-            res = getattr(self._proxy, 'on_change_with_%s' % field)(args,
+            result = getattr(self._proxy, 'on_change_with')(values,
+                list(to_change), context)
+            for field, value in result.items():
+                self._on_change_set(field, value)
+        for field in later:
+            definition = self._fields[field]
+            values = self._on_change_args(definition['on_change_with'])
+            context = self._config.context
+            result = getattr(self._proxy, 'on_change_with_%s' % field)(values,
                     context)
-            self._on_change_set(field, res)
+            self._on_change_set(field, result)
         if self._parent:
             self._parent._changed.add(self._parent_field_name)
             self._parent._on_change(self._parent_field_name)
@@ -767,67 +932,82 @@ class Model(object):
 class Wizard(object):
     'Wizard class for Tryton wizards'
 
-    def __init__(self, name, models=None, config=None, context=None):
+    def __init__(self, name, models=None, action=None, config=None,
+            context=None):
         if models:
             assert len(set(type(x) for x in models)) == 1
         super(Wizard, self).__init__()
         self.name = name
-        self.state = None
-        self.states = ['init']
         self.form = None
+        self.form_state = None
         self._config = config or proteus.config.get_config()
         self._context = context or {}
         self._proxy = self._config.get_proxy(name, type='wizard')
-        self.id = self._proxy.create(self._config.context)
-        if models:
-            self.datas = {
-                    'model': models[0].__class__.__name__,
-                    'id': models[0].id,
-                    'ids': [model.id for model in models]
-                    }
-        else:
-            self.datas = {}
-        self.execute('init')
+        result = self._proxy.create(self._config.context)
+        self.session_id, self.start_state, self.end_state = result
+        self.states = [self.start_state]
+        self.models = models
+        self.action = action
+        self.execute(self.start_state)
 
     def execute(self, state):
         assert state in self.states
 
-        if 'form' not in self.datas:
-            self.datas['form'] = {}
-
-        if self.form:
-            self.datas['form'].update(self.form._get_values())
-
         self.state = state
-        while self.state != 'end':
+        while self.state != self.end_state:
             ctx = self._context.copy()
             ctx.update(self._config.context)
-            ctx['active_id'] = self.datas.get('id')
-            ctx['active_ids'] = self.datas.get('ids')
+            if self.models:
+                ctx['active_id'] = self.models[0].id
+                ctx['active_ids'] = [model.id for model in self.models]
+                ctx['active_model'] = self.models[0].__class__.__name__
+            else:
+                ctx['active_id'] = None
+                ctx['active_ids'] = None
+                ctx['active_model'] = None
+            if self.action:
+                ctx['action_id'] = self.action.id
+            else:
+                ctx['action_id'] = None
 
-            res = self._proxy.execute(self.id, self.datas, self.state, ctx)
-            if not res:
-                break
+            if self.form:
+                # Filter only modified values
+                data = {self.form_state:
+                    dict((k, v) for k, v in
+                        self.form._get_on_change_values().iteritems()
+                        if k in self.form._values)}
+            else:
+                data = {}
 
-            if 'datas' in res:
-                self.datas['form'] = res['datas']
-            elif res['type'] == 'form':
-                self.datas['form'] = {}
+            result = self._proxy.execute(self.session_id, data, self.state,
+                ctx)
 
-            if res['type'] == 'form':
-                self.states = [x[0] for x in res['state']]
-                # XXX set context
-                self.form = Model.get(res['object'])()
-                self.form._default_set(self.datas['form'])
+            if 'view' in result:
+                view = result['view']
+                self.form = Model.get(view['fields_view']['model'])()
+                self.form._default_set(view['defaults'])
+                self.states = [b['state'] for b in view['buttons']]
+                self.form_state = view['state']
                 return
-            elif res['type'] == 'action':
-                # TODO run action
-                self.state = res['state']
-            elif res['type'] == 'print':
-                # TODO run print
-                self.state = res['state']
-            elif res['type'] == 'state':
-                self.state = res['state']
+            else:
+                self.state = self.end_state
 
-        if self.state == 'end':
-            self._proxy.delete(self.id, self._config.context)
+            if 'actions' in result:
+                pass  # TODO
+
+        if self.state == self.end_state:
+            self._proxy.delete(self.session_id, self._config.context)
+
+
+class Report(object):
+    'Report class for Tryton reports'
+
+    def __init__(self, name, config=None, context=None):
+        super(Report, self).__init__()
+        self.name = name
+        self._config = config or proteus.config.get_config()
+        self._context = context or {}
+        self._proxy = self._config.get_proxy(name, type='report')
+
+    def execute(self, models, data):
+        return self._proxy.execute([m.id for m in models], data, self._context)
