@@ -1,13 +1,10 @@
 #This file is part of Tryton.  The COPYRIGHT file at the top level of
 #this repository contains the full copyright notices and license terms.
-from unittest import TestCase
-from proteus import config, Model
+from proteus import Model
+from .common import ProteusTestCase
 
 
-class TestModel(TestCase):
-
-    def setUp(self):
-        config.set_trytond(database_type='sqlite')
+class TestModel(ProteusTestCase):
 
     def test_class_cache(self):
         User1 = Model.get('res.user')
@@ -39,7 +36,7 @@ class TestModel(TestCase):
         admin.create_uid = admin
         admin.create_uid = None
 
-        test = User(write_uid=False)
+        User(write_uid=None)
 
     def test_one2many(self):
         Group = Model.get('res.group')
@@ -241,7 +238,7 @@ class TestModel(TestCase):
         self.assertNotEqual(admin1, None)
         self.assertNotEqual(admin1, False)
 
-        self.failUnlessRaises(NotImplementedError, lambda: admin1 == 1)
+        self.assertRaises(NotImplementedError, lambda: admin1 == 1)
 
     def test_default_set(self):
         User = Model.get('res.user')
@@ -276,7 +273,18 @@ class TestModel(TestCase):
         test.name = 'Test delete'
         test.login = 'test delete'
         test.save()
-        self.assert_(test.delete())
+        test.delete()
+
+    def test_duplicate(self):
+        User = Model.get('res.user')
+        test = User()
+        test.name = 'Test duplicate'
+        test.login = 'test duplicate'
+        test.save()
+        copy, = User.duplicate([test], {'name': 'Test copy'})
+        self.assertEqual(copy.name, 'Test copy')
+        self.assertEqual(copy.login, 'test duplicate (copy)')
+        self.assertNotEqual(copy.id, test.id)
 
     def test_on_change(self):
         Trigger = Model.get('ir.trigger')
@@ -289,6 +297,14 @@ class TestModel(TestCase):
         trigger.on_create = True
         self.assertEqual(trigger.on_time, False)
 
+    def test_on_change_with(self):
+        Attachment = Model.get('ir.attachment')
+
+        attachment = Attachment()
+
+        attachment.description = 'Test'
+        self.assertEqual(attachment.summary, 'Test')
+
     def test_on_change_set(self):
         User = Model.get('res.user')
         Group = Model.get('res.group')
@@ -303,9 +319,9 @@ class TestModel(TestCase):
         test._on_change_set('groups', {'remove': [group_ids[0]]})
         self.assertEqual([x.id for x in test.groups], group_ids[1:])
 
-        test._on_change_set('groups', {'add': [{
-            'name': 'Bar',
-            }]})
+        test._on_change_set('groups', {'add': [(-1, {
+                            'name': 'Bar',
+                            })]})
         self.assert_([x for x in test.groups if x.name == 'Bar'])
 
         test.groups.extend(Group.find())
